# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs

DEFAULT_CRS = ccrs.PlateCarree()


CRS_MAPPING = {
    "EPSG:4326": ccrs.PlateCarree,
    "cylindrical": ccrs.PlateCarree,
}


def parse(crs):
    """
    Convert a string or dictionary representation of a CRS into a cartopy CRS.

    Parameters
    ----------
    crs : str, dict or cartopy.crs.CRS
        Some representation of a CRS to be parsed and converted into a cartopy
        CRS.
        If a string, must be the name of a cartopy CRS.
        If a dictionary, must include a "name" key matching the name of a
        cartopy CRS, plus and keyword arguments to be passed to the CRS
        constructor.

    Example
    -------
    >>> earthkit.maps.domains.parse_crs("PlateCarree")
    <Derived Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to ...>
    Name: unknown
    Axis Info [cartesian]:
    - E[east]: Easting (unknown)
    - N[north]: Northing (unknown)
    - h[up]: Ellipsoidal height (metre)
    Area of Use:
    - undefined
    Coordinate Operation:
    - name: unknown
    - method: Equidistant Cylindrical
    Datum: unknown
    - Ellipsoid: WGS 84
    - Prime Meridian: Greenwich
    >>> earthkit.maps.domains.parse_crs(
    ...     {"name": "PlateCarree", "central_longitude": 50}
    ... )
    <Derived Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=50 +to_ ...>
    Name: unknown
    Axis Info [cartesian]:
    - E[east]: Easting (unknown)
    - N[north]: Northing (unknown)
    - h[up]: Ellipsoidal height (metre)
    Area of Use:
    - undefined
    Coordinate Operation:
    - name: unknown
    - method: Equidistant Cylindrical
    Datum: unknown
    - Ellipsoid: WGS 84
    - Prime Meridian: Greenwich

    Returns
    -------
    cartopy.crs.CRS
    """
    if crs is None:
        crs = DEFAULT_CRS

    if not isinstance(crs, ccrs.CRS):
        if isinstance(crs, dict):
            crs = from_dict(crs)
        else:
            if crs in CRS_MAPPING:
                crs = CRS_MAPPING[crs]()
            elif crs.upper().startswith("EPSG"):
                crs = crs.upper().lstrip("EPSG:")
                crs = ccrs.epsg(crs)
            else:
                crs = from_string(crs)

    return crs


def from_dict(kwargs):
    """
    Convert a dictionary representation of a CRS into a cartopy CRS.

    Parameters
    ----------
    crs : dict
        A dictionary representation of a CRS to be parsed and converted into a
        cartopy CRS. Must include a "name" key matching the name of a cartopy
        CRS, plus and keyword arguments to be passed to the CRS constructor.

    Example
    -------
    >>> earthkit.maps.domains.parse_crs(
    ...     {"name": "PlateCarree", "central_longitude": 50}
    ... )
    <Derived Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=50 +to_ ...>
    Name: unknown
    Axis Info [cartesian]:
    - E[east]: Easting (unknown)
    - N[north]: Northing (unknown)
    - h[up]: Ellipsoidal height (metre)
    Area of Use:
    - undefined
    Coordinate Operation:
    - name: unknown
    - method: Equidistant Cylindrical
    Datum: unknown
    - Ellipsoid: WGS 84
    - Prime Meridian: Greenwich

    Returns
    -------
    cartopy.crs.CRS
    """
    crs = from_string(kwargs["name"]).__class__
    return crs(**{k: v for k, v in kwargs.items() if k != "name"})


def from_string(string):
    """
    Convert a string name of a CRS into a cartopy CRS.

    Parameters
    ----------
    crs : dict
        A string matching the name of a cartopy CRS.

    Example
    -------
    >>> earthkit.maps.domains.parse_crs("PlateCarree")
    <Derived Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to ...>
    Name: unknown
    Axis Info [cartesian]:
    - E[east]: Easting (unknown)
    - N[north]: Northing (unknown)
    - h[up]: Ellipsoidal height (metre)
    Area of Use:
    - undefined
    Coordinate Operation:
    - name: unknown
    - method: Equidistant Cylindrical
    Datum: unknown
    - Ellipsoid: WGS 84
    - Prime Meridian: Greenwich

    Returns
    -------
    cartopy.crs.CRS
    """
    try:
        crs = getattr(ccrs, string)()
    except AttributeError:
        raise ValueError(f"cartopy has no CRS named '{string}'")
    return crs



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import cartopy.io.shapereader as shpreader
import earthkit.data
import matplotlib.pyplot as plt

from earthkit.maps import domains
from earthkit.maps.domains import natural_earth
from earthkit.maps.layers import metadata
from earthkit.maps.layers.layers import Layer, LayerFormatter
from earthkit.maps.schemas import schema
from earthkit.maps.styles.auto import suggest_style


class SubplotFormatter(metadata.BaseFormatter):
    def __init__(self, subplot, unique=True):
        self.subplot = subplot
        self.unique = unique
        self._layer_index = None

    def convert_field(self, value, conversion):
        f = super().convert_field
        if isinstance(value, list):
            return [f(v, conversion) for v in value]
        else:
            return f(value, conversion)

    def format_key(self, key):
        values = [
            LayerFormatter(layer).format_key(key) for layer in self.subplot.layers
        ]
        return values

    def format_field(self, value, format_spec):
        f = super().format_field
        if isinstance(value, list):
            values = [f(v, format_spec) for v in value]
            if self._layer_index is not None:
                value = values[self._layer_index]
                self._layer_index = None
            else:
                if self.unique:
                    values = list(dict.fromkeys(values))
                value = metadata.list_to_human(values)
        return value


class Subplot:
    @classmethod
    def from_data(cls, superplot, data, *args, domain=None, crs=None, **kwargs):
        if not isinstance(data, earthkit.data.core.Base):
            data = earthkit.data.from_object(data)

        if domain is None and crs is None:
            try:
                crs = data.projection().to_cartopy_crs()
            except AttributeError:
                pass
        return cls(superplot, *args, domain=domain, crs=crs, **kwargs)

    def __init__(self, superplot, *args, domain=None, crs=None, **kwargs):
        self.domain = domains.parse(domain, crs)

        self.ax = superplot.fig.add_subplot(*args, projection=self.domain.crs, **kwargs)

        if self.domain.bounds is not None:
            self.ax.set_extent(self.domain.bounds, crs=self.domain.crs)

        self.superplot = superplot
        self.layers = []

    @property
    def fig(self):
        return self.superplot.fig

    @property
    def distinct_legend_layers(self):
        """Group layers by style."""
        legend_layers = []
        for layer in self.layers:
            for legend_layer in legend_layers:
                if legend_layer.style == layer.style:
                    break
            else:
                legend_layers.append(layer)
        return legend_layers

    def gridded_scalar(method):
        def wrapper(self, data, x=None, y=None, transform=None, style=None, **kwargs):
            # - TEMPORARY: in the future all "fields" will be "fieldlists" -
            if isinstance(data, earthkit.data.core.Base):
                try:
                    data = data[0]
                except (ValueError, TypeError):
                    pass
            # --------------------------------------------------------------

            if x is not None and y is not None:
                if transform is None:
                    raise ValueError(
                        "you must pass a 'transform' when plotting manually "
                        "with x and y coordinates"
                    )
                values = data
            else:
                if not isinstance(data, earthkit.data.core.Base):
                    data = earthkit.data.from_object(data)
                x, y, values = extract_scalar(data, self.domain)
                if transform is None:
                    transform = data.projection().to_cartopy_crs()

            try:
                source_units = data.metadata("units")
            except AttributeError:
                source_units = None

            if style is None:
                style_units = kwargs.pop("units", source_units)
                style = suggest_style(data, units=style_units)

            values = style.convert_units(values, source_units)

            if "transform_first" not in kwargs:
                kwargs["transform_first"] = self._can_transform_first(method)

            mappable = method(
                self, x, y, values, style=style, transform=transform, **kwargs
            )

            layer = Layer(data, mappable, self, style=style)
            self.layers.append(layer)

            return layer

        return wrapper

    def _can_transform_first(self, method):
        if self.domain._can_transform_first and method.__name__ != "pcolormesh":
            return True
        else:
            return False

    @gridded_scalar
    def plot(self, *args, style=None, **kwargs):
        try:
            return style.plot(self.ax, *args, **kwargs)
        except NotImplementedError:
            return style.contourf(self.ax, *args, **kwargs)

    @gridded_scalar
    def contourf(self, *args, style=None, **kwargs):
        return style.contourf(self.ax, *args, **kwargs)

    @gridded_scalar
    def contour(self, *args, style=None, **kwargs):
        return style.contour(self.ax, *args, **kwargs)

    shaded_contour = contourf

    @gridded_scalar
    def pcolormesh(self, *args, style=None, **kwargs):
        return style.pcolormesh(self.ax, *args, **kwargs)

    @gridded_scalar
    def scatter(self, *args, style=None, **kwargs):
        return style.scatter(self.ax, *args, **kwargs)

    @schema.coastlines.apply()
    def coastlines(self, *args, resolution="auto", **kwargs):
        """Add coastal outlines from the Natural Earth “coastline” collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        return self.ax.coastlines(*args, resolution=resolution, **kwargs)

    @schema.borders.apply()
    def borders(
        self, *args, resolution="auto", labels=False, label_kwargs=None, **kwargs
    ):
        """Add country boundary polygons from Natural Earth.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.BORDERS
        else:
            feature = cfeature.NaturalEarthFeature(
                "cultural", "admin_0_countries", resolution
            )
        if labels:
            label_kwargs = label_kwargs or dict()
            label_kwargs = {
                **dict(
                    ha="center",
                    va="center",
                    bbox=dict(
                        boxstyle="round",
                        ec=(0.2, 0.2, 0.2, 0),
                        fc=(0.3, 0.3, 0.3),
                    ),
                    fontsize=8,
                    weight="bold",
                    color=(0.95, 0.95, 0.95),
                    clip_on=True,
                    clip_box=self.ax.bbox,
                    transform=ccrs.Geodetic(),
                ),
                **label_kwargs,
            }

            label_key = labels if isinstance(labels, str) else "ISO_A2_EH"

            resolution = "110m" if resolution == "auto" else resolution
            shpfilename = shpreader.natural_earth(
                resolution=resolution, category="cultural", name="admin_0_countries"
            )
            reader = shpreader.Reader(shpfilename)
            for record in reader.records():
                name = record.attributes[label_key]
                x = record.attributes["LABEL_X"]
                y = record.attributes["LABEL_Y"]
                self.ax.text(x, y, name, **label_kwargs)
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.land.apply()
    def land(self, *args, resolution="auto", **kwargs):
        """Add land polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.LAND
        else:
            feature = cfeature.NaturalEarthFeature("physical", "land", resolution)
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.land.apply()
    def ocean(self, *args, resolution="auto", **kwargs):
        """Add ocean polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.OCEAN
        else:
            feature = cfeature.NaturalEarthFeature("physical", "land", resolution)
        return self.ax.add_feature(feature, *args, **kwargs)

    def stock_img(self, *args, **kwargs):
        self.ax.stock_img(*args, **kwargs)

    @schema.gridlines.apply()
    def gridlines(self, *args, **kwargs):
        """Add gridlines to the map."""
        self._gridlines = self.ax.gridlines(*args, **kwargs)
        return self._gridlines

    @schema.legend.apply()
    def legend(self, *args, **kwargs):
        legends = []
        for layer in self.distinct_legend_layers:
            if layer.style is not None:
                legend = layer.style.legend(
                    self.fig,
                    layer,
                    *args,
                    ax=layer.axes,
                    **kwargs,
                )
            legends.append(legend)
        return legends

    @property
    def _default_title_template(self):
        templates = [layer._default_title_template for layer in self.layers]
        if len(set(templates)) == 1:
            template = templates[0]
        else:
            title_parts = []
            for i, template in enumerate(templates):
                keys = [k for _, k, _, _ in SubplotFormatter().parse(template)]
                for key in set(keys):
                    template = template.replace("{" + key, "{" + key + f"!{i}")
                title_parts.append(template)
            template = metadata.list_to_human(title_parts)
        return template

    @schema.title.apply()
    def title(self, label=None, unique=True, wrap=True, **kwargs):
        if label is None:
            label = self._default_title_template
        label = self.format_string(label, unique)
        plt.sca(self.ax)
        return plt.title(label, wrap=wrap, **kwargs)

    def format_string(self, string, unique=True):
        return SubplotFormatter(self, unique=unique).format(string)


def extract_scalar(data, domain):
    try:
        data = data[0]
    except (ValueError, TypeError):
        data = data

    values, points = domain.bbox(data)

    y = points["y"]
    x = points["x"]
    # if data.projection().CF_GRID_MAPPING_NAME == "latitude_longitude":
    #     x[x > 180] -= 360

    return x, y, values


def extract_vector(data, domain):
    x, y, u_values = extract_scalar(data[0], domain)
    _, _, v_values = extract_scalar(data[1], domain)
    return x, y, (u_values, v_values)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.io.shapereader as shpreader

from earthkit.maps import domains

RESOLUTIONS = {
    "low": "110m",
    "medium": "50m",
    "high": "10m",
}


class NaturalEarthDomain:
    """Class for building map domains and CRS based on a Natural Earth shape."""

    NATURAL_EARTH_SOURCES = {
        "admin_0_map_units": "NAME_LONG",
        "admin_0_countries": "NAME_LONG",
        "admin_1_states_provinces": "name_en",
    }

    def __init__(self, domain_name, crs=None):
        self._domain_name = domain_name
        self._record = None
        self._source = None
        self._crs = crs

    @property
    def domain_name(self):
        return self.record.attributes.get(self.NATURAL_EARTH_SOURCES[self._source])

    @property
    def record(self):
        if self._record is None:
            for source, attribute in self.NATURAL_EARTH_SOURCES.items():
                shpfilename = shpreader.natural_earth(
                    resolution="110m", category="cultural", name=source
                )
                reader = shpreader.Reader(shpfilename)
                for record in reader.records():
                    name = record.attributes.get(attribute) or ""
                    name = name.replace("\x00", "")
                    if name.lower() == self._domain_name.lower():
                        break
                else:
                    continue
                break
            else:
                raise ValueError(
                    f"No country or state named '{self._domain_name}' found in "
                    f"Natural Earth's shapefiles"
                )
            self._record = record
            self._source = source

        return self._record

    @property
    def geometry(self):
        return self.record.geometry

    @property
    def crs(self):
        if self._crs is None:
            bounds = domains.bounds.from_geometry(self.geometry)
            self._crs = domains.optimal.crs_from_bounds(bounds)

        return self._crs

    @property
    def bounds(self, pad=0.1):
        crs_bounds = domains.bounds.from_geometry(self.geometry, crs=self.crs)

        if pad:
            x_offset = (crs_bounds[1] - crs_bounds[0]) * pad
            y_offset = (crs_bounds[3] - crs_bounds[2]) * pad
            offset = min(x_offset, y_offset)
            crs_bounds = [
                crs_bounds[0] - offset,
                crs_bounds[1] + offset,
                crs_bounds[2] - offset,
                crs_bounds[3] + offset,
            ]

        return crs_bounds



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from earthkit.maps import styles

TEMPERATURE_AT_2M_IN_CELSIUS = styles.Contour(
    colors="turbo",
    extend="both",
    levels=range(-40, 41, 5),
    units="celsius",
)

MEAN_SEA_LEVEL_PRESSURE_IN_HPA = styles.Contour(
    line_colors="#33334d",
    linewidths=[0.5, 0.5, 0.5, 1],
    labels=True,
    level_step=4,
    units="hPa",
    legend_type=None,
)


WIND_GUST_AT_10M_OVER_6_HOURS_IN_METRES_PER_SECOND = styles.Contour(
    colors=["#73A3FF", "#0080FF", "#6CA632", "#FF8000", "#FF0A00", "#7B12B3"],
    line_colors="navy",
    levels=range(10, 45, 5),
    extend="max",
)


EFI_WIND_GUST_AT_10M = styles.Hatched(
    levels=[0.6, 0.8, 1.0],
    hatches=["." * 5, "o" * 5],
    legend_type="disjoint",
    colors="magenta",
)


EFI_TOTAL_PRECIPITATION = styles.Hatched(
    levels=[0.6, 0.8, 1.0],
    hatches=["." * 5, "o" * 5],
    colors="green",
)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Import and path setup ---------------------------------------------------

import os
import sys

# import earthkit.maps

sys.path.insert(0, os.path.abspath("../"))

# -- Project information -----------------------------------------------------

project = "earthkit-maps"
copyright = "2023, European Centre for Medium Range Weather Forecasts"
author = "European Centre for Medium Range Weather Forecasts"
version = "0.0.0"  # earthkit.maps.__version__
release = "0.0.0"  # earthkit.maps.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # "autoapi.extension",
    # "myst_parser",
    # "sphinx.ext.autodoc",
    # "sphinx.ext.napoleon",
]

# autodoc configuration
autodoc_typehints = "none"

# autoapi configuration
# autoapi_dirs = ["../earthkit/maps"]
# autoapi_ignore = ["*/version.py"]
# autoapi_options = [
#     "members",
#     "inherited-members",
#     "undoc-members",
#     "show-inheritance",
#     "show-module-summary",
#     "imported-members",
# ]
# autoapi_root = "_api"

# napoleon configuration
# napoleon_google_docstring = False
# napoleon_numpy_docstring = True
# napoleon_preprocess_types = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import numpy as np
from earthkit.data.utils.bbox import BoundingBox

CYCLIC_SYSTEMS = ["PlateCarree", "Mercator"]


def from_geometry(geom, crs=None):
    """
    Get a bounding box around a geometry on a given coordinate reference system.

    Parameters
    ----------
    geom : shapely.geometry
        A shapely geometry around which the bounding box should be drawn.
    crs : cartopy.crs.CRS
        The coordinate reference system in which to generate a new bounding box
        which entriely contains the given geometry.

    Returns
    -------
    list
    """
    try:
        geoms = geom.geoms
    except AttributeError:
        geoms = [geom]
    iterator = list(geoms)

    x = np.concatenate([g.boundary.xy[0] for g in iterator])
    y = np.concatenate([g.boundary.xy[1] for g in iterator])

    if crs is not None:
        xy = crs.transform_points(x=x, y=y, src_crs=ccrs.PlateCarree())
        x = [point[0] for point in xy]
        y = [point[1] for point in xy]
    elif any(lon in (min(x), max(x)) for lon in (-180, 180)):
        x = [v % 360 for v in x]

    xmin = min(x)
    xmax = max(x)

    ymin = min(y)
    ymax = max(y)

    return [xmin, xmax, ymin, ymax]


def from_bbox(extents, crs, src_crs=None):
    """
    Get a bounding box which entirely contains another box from a different CRS.

    Parameters
    ----------
    extents : list
        The x and y extents of the bounding box on the `src_crs`, given as
        `[min_x, max_x, min_y, max_y]`.
    crs : cartopy.crs.CRS
        The coordinate reference system in which to generate a new bounding box
        which entriely contains the given extents.
    src_crs : cartopy.crs.CRS (optional)
        The coordinate reference system on which the extents are defined. If not
        passed, an equirectangular CRS will be assumed, with extents defined as
        latitudes and longitudes.

    Returns
    -------
    list
    """
    # TODO: Implement bbox conversion for stereographic projections
    if isinstance(src_crs, ccrs.Stereographic):
        raise NotImplementedError(f"cannot convert bbox for CRS {src_crs}")

    if src_crs is None:
        src_crs = ccrs.PlateCarree()

    if isinstance(extents, BoundingBox):
        extents = [extents.west, extents.east, extents.south, extents.north]

    ll_grid = crs.__class__.__name__ in CYCLIC_SYSTEMS

    min_lon, max_lon, min_lat, max_lat = extents
    mid_lon = max_lon - (max_lon - min_lon) / 2

    corners = [
        crs.transform_point(min_lon, min_lat, src_crs),
        crs.transform_point(min_lon, max_lat, src_crs),
        crs.transform_point(max_lon, max_lat, src_crs),
        crs.transform_point(max_lon, min_lat, src_crs),
    ]

    min_x = min([corner[0] for corner in corners[:2]])
    max_x = max([corner[0] for corner in corners[2:4]])
    min_y = min([corner[1] for corner in [corners[0], corners[-1]]])
    max_y = max([corner[1] for corner in corners[1:3]])

    if ll_grid and (
        (abs(corners[2][0] - corners[3][0]) > 180)
        or (abs(corners[0][0] - corners[1][0]) > 180)
    ):
        min_x = min([corner[0] % 360 for corner in corners[:2]])
        max_x = max([corner[0] % 360 for corner in corners[2:4]])

    mid_bottom = crs.transform_point(mid_lon, min_lat, src_crs)
    mid_top = crs.transform_point(mid_lon, max_lat, src_crs)

    min_y = min(min_y, mid_bottom[1])
    max_y = max(max_y, mid_top[1])

    if min_x > max_x and ll_grid:
        min_x %= 360
        max_x %= 360

    return [min_x, max_x, min_y, max_y]



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from earthkit.maps.layers import metadata


class LayerFormatter(metadata.BaseFormatter):
    def __init__(self, layer):
        self.layer = layer

    def format_keys(self, format_string, kwargs):
        keys = (i[1] for i in self.parse(format_string) if i[1] is not None)
        for key in keys:
            kwargs[key] = self.format_key(key)
        return kwargs

    def format_key(self, key):
        if key in self.SUBPLOT_ATTRIBUTES:
            value = getattr(self.layer.subplot, self.SUBPLOT_ATTRIBUTES[key])
        elif key in self.STYLE_ATTRIBUTES and self.layer.style is not None:
            value = getattr(self.layer.style, self.STYLE_ATTRIBUTES[key])
            if value is None:
                value = metadata.get_metadata(self.layer, key)
                if key == "units":
                    value = metadata.format_units(value)
        else:
            value = metadata.get_metadata(self.layer, key)
        return value


class Layer:
    def __init__(self, data, mappable, subplot, style=None):
        self.data = data
        self.mappable = mappable
        self.subplot = subplot
        self.style = style

    @property
    def fig(self):
        return self.subplot.fig

    def legend(self, *args, **kwargs):
        if self.style is not None:
            return self.style.legend(
                self.fig,
                self,
                *args,
                ax=self.axes,
                **kwargs,
            )

    @property
    def ax(self):
        return self.subplot.ax

    @property
    def axes(self):
        return [self.ax]

    def format_string(self, string):
        return LayerFormatter(self).format(string)

    @property
    def _default_title_template(self):
        if self.data.metadata("type") == "an":
            template = metadata.DEFAULT_ANALYSIS_TITLE
        else:
            template = metadata.DEFAULT_FORECAST_TITLE
        return template


class MultiplotLayer:
    def __init__(self, layers):
        self.layers = layers

    @property
    def subplots(self):
        return [layer.subplot for layer in self.layers]

    @property
    def fig(self):
        return self.subplots[0].fig

    @property
    def axes(self):
        return [subplot.ax for subplot in self.subplots]

    @property
    def style(self):
        return self.layers[0].style

    def legend(self, *args, **kwargs):
        if self.style is not None:
            return self.style.legend(
                self.fig,
                self,
                *args,
                ax=self.axes,
                **kwargs,
            )

    @property
    def mappable(self):
        return self.layers[0].mappable

    def format_string(self, string, unique=True):
        results = [layer.format_string(string) for layer in self.layers]

        if unique:
            results = list(dict.fromkeys(results))

        result = metadata.list_to_human(results)
        return result



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import earthkit.data
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from cf_units import Unit
from matplotlib.colors import BoundaryNorm, LinearSegmentedColormap
from matplotlib.patches import Rectangle

from earthkit.maps.layers import metadata
from earthkit.maps.schemas import schema

DEFAULT_LEGEND_LABEL = "{variable_name} ({units})"


def auto_range(data, n_levels=8):

    if isinstance(data, earthkit.data.core.Base):
        data = data.to_numpy()
    min_value = np.nanmin(data)
    max_value = np.nanmax(data)

    data_range = max_value - min_value

    initial_bin = data_range / n_levels

    magnitude = 10 ** (math.floor(math.log(initial_bin, 10)))
    bin_width = initial_bin - (initial_bin % -magnitude)

    start = min_value - (min_value % magnitude)

    levels = list(np.arange(start, start + bin_width * n_levels + bin_width, bin_width))

    while levels[-2] > max_value:
        levels = levels[:-1]

    return levels


def step_range(data, step, multiple=None):

    if multiple is None:
        multiple = step

    if isinstance(data, earthkit.data.core.Base):
        data = data.to_numpy()
    min_value = np.nanmin(data)
    max_value = np.nanmax(data)

    min_value = min_value - (min_value % multiple)
    max_value = int(math.ceil(max_value / multiple)) * multiple

    levels = np.arange(min_value, max_value + step, step)
    return levels


def expand_colors(colors, levels):
    if isinstance(colors, (list, tuple)) and len(colors) == 1:
        colors *= len(levels) - 1
    if isinstance(colors, str):
        try:
            cmap = mpl.cm.get_cmap(colors)
        except ValueError:
            colors = [colors] * (len(levels) - 1)
        else:
            colors = [mpl.cm.get_cmap(cmap)(i) for i in np.linspace(0, 1, len(levels))]
    return colors


class Style:
    def __init__(
        self,
        colors="auto",
        levels=None,
        level_step=None,
        level_multiple=None,
        units=None,
        units_override=None,
        normalize=True,
        legend_type="colorbar",
        categories=None,
        conversion=None,
        **kwargs,
    ):
        if colors == "auto":
            self._colors = schema.cmap
        else:
            self._colors = colors

        self._levels = levels
        self._level_step = level_step
        self._level_multiple = level_multiple

        self.legend_type = legend_type

        self._units = units
        self._units_override = units_override
        self.normalize = normalize
        self.kwargs = kwargs

        self.conversion = conversion

        self._categories = categories

    @property
    def units(self):
        if self._units_override is not None:
            return metadata.format_units(self._units_override)
        elif self._units is not None:
            return metadata.format_units(self._units)

    def convert_units(self, values, source_units):
        if self.conversion is not None:
            values = self.conversion(values)

        if self._units is None:
            return values
        return Unit(source_units).convert(values, self._units)

    def levels(self, data):
        if self._levels is None:
            if self._level_step is None:
                return auto_range(data)
            else:
                return step_range(data, self._level_step, self._level_multiple)
        return self._levels

    def get_levels(self, layer):
        return layer.mappable.norm.boundaries

    def to_kwargs(self, data):
        levels = self.levels(data)

        colors = expand_colors(self._colors, levels)

        cmap = LinearSegmentedColormap.from_list(name="", colors=colors, N=len(colors))

        norm = None
        if self.normalize:
            norm = BoundaryNorm(levels, cmap.N)

        return {
            **{"cmap": cmap, "norm": norm, "levels": levels},
            **self.kwargs,
        }

    def to_contourf_kwargs(self, data):
        kwargs = self.to_kwargs(data)
        kwargs.pop("linewidths", None)
        return kwargs

    def to_contour_kwargs(self, data):
        return self.to_kwargs(data)

    def to_pcolormesh_kwargs(self, data):
        kwargs = self.to_kwargs(data)
        kwargs.pop("levels", None)
        kwargs.pop("transform_first", None)
        kwargs.pop("extend", None)
        return kwargs

    def to_scatter_kwargs(self, data):
        kwargs = self.to_kwargs(data)
        kwargs.pop("levels", None)
        return kwargs

    def plot(self, *args, **kwargs):
        raise NotImplementedError("Generic styles cannot be used with 'plot'")

    def contourf(self, ax, x, y, values, *args, **kwargs):
        kwargs = {**self.to_contourf_kwargs(values), **kwargs}
        return ax.contourf(x, y, values, *args, **kwargs)

    def contour(self, ax, x, y, values, *args, **kwargs):
        kwargs = {**self.to_contour_kwargs(values), **kwargs}
        return ax.contour(x, y, values, *args, **kwargs)

    def pcolormesh(self, ax, x, y, values, *args, **kwargs):
        kwargs.pop("transform_first", None)
        kwargs = {**self.to_pcolormesh_kwargs(values), **kwargs}
        return ax.pcolormesh(x, y, values, *args, **kwargs)

    def scatter(self, ax, x, y, values, s=3, *args, **kwargs):
        kwargs.pop("transform_first", None)
        kwargs = {**self.to_scatter_kwargs(values), **kwargs}
        return ax.scatter(x, y, c=values, s=s, *args, **kwargs)

    def colorbar(self, fig, layer, *args, shrink=0.8, aspect=35, **kwargs):
        label = kwargs.pop("label", DEFAULT_LEGEND_LABEL)
        label = layer.format_string(label)

        levels = self.get_levels(layer)
        if len(np.unique(np.ediff1d(levels))) > 1:
            kwargs["ticks"] = kwargs.get("ticks", levels)
        kwargs["format"] = kwargs.get("format", lambda x, _: f"{x:g}")

        colorbar = fig.colorbar(
            layer.mappable,
            *args,
            label=label,
            shrink=shrink,
            aspect=aspect,
            **kwargs,
        )
        colorbar.ax.minorticks_off()

        return colorbar

    def disjoint_legend(
        self, fig, layer, ax, *args, location="bottom", frameon=False, **kwargs
    ):
        kwargs.pop("format")

        label = kwargs.pop("label", DEFAULT_LEGEND_LABEL)
        label = layer.format_string(label)

        source = ax[0] if len(ax) == 1 else fig

        location_kwargs = {
            "bottom": {
                "loc": "upper center",
                "bbox_to_anchor": (0.5, -0.05),
            },
            "top": {
                "loc": "lower center",
                "bbox_to_anchor": (0.5, 1.0),
            },
            "left": {
                "loc": "upper right",
                "bbox_to_anchor": (-0.05, 1.0),
            },
            "right": {
                "loc": "upper left",
                "bbox_to_anchor": (1.05, 1.0),
            },
            "top left": {
                "loc": "lower center",
                "bbox_to_anchor": (0.25, 1),
            },
            "top right": {
                "loc": "lower center",
                "bbox_to_anchor": (0.75, 1),
            },
        }[location]

        artists, labels = layer.mappable.legend_elements()

        labels = kwargs.pop("labels", self._categories) or labels

        legend = source.legend(
            artists,
            labels,
            *args,
            title=label,
            frameon=frameon,
            **{**location_kwargs, **kwargs},
        )

        # Matplotlib removes legends when a new legend is plotted, so we have
        # to manually re-add them...
        if hasattr(fig, "_previous_legend"):
            fig.add_artist(fig._previous_legend)
        fig._previous_legend = legend

        return legend

    disjoint = disjoint_legend

    def histogram_legend(self, fig, layer, *args, aspect=20, **kwargs):
        label = kwargs.pop("label", DEFAULT_LEGEND_LABEL)
        label = layer.format_string(label)

        colorbar = self.colorbar(fig, layer, *args, aspect=aspect, label="", **kwargs)
        colorbar.outline.set_linestyle(":")
        colorbar.outline.set_alpha(0.5)
        cax = colorbar.ax
        cax.clear()

        levels = self.get_levels(layer)

        location = kwargs.get("location")
        orientation = "vertical" if location in ("top", "bottom") else "horizontal"

        if location in ("top", "bottom"):
            orientation = "vertical"
            cax.xaxis.set_label_position("bottom")
            cax.xaxis.set_ticks_position("bottom")
            cax.set_xlabel(label)
        else:
            orientation = "horizontal"
            cax.yaxis.set_label_position("left")
            cax.yaxis.set_ticks_position("left")
            cax.set_ylabel(label)

        n, bins, patches = cax.hist(
            layer.layers[0].data,
            orientation=orientation,
            bins=levels,
        )

        patch_height = -max(n) / 5
        colors = [layer.mappable.cmap(i) for i in np.linspace(0, 1, len(levels))]

        if orientation == "vertical":
            cax.set_xlim([levels[0], levels[-1]])
            cax.set_ylim([patch_height, None])
            for i in range(len(levels) - 1):
                cax.add_patch(
                    Rectangle(
                        (levels[i], patch_height),
                        levels[i + 1] - levels[i],
                        -patch_height,
                        facecolor=colors[i],
                        edgecolor="black",
                        linewidth=0.5,
                    )
                )
        else:
            cax.set_ylim([levels[0], levels[-1]])
            cax.set_xlim([patch_height, None])
            for i in range(len(levels) - 1):
                cax.add_patch(
                    Rectangle(
                        (patch_height, levels[i]),
                        -patch_height,
                        levels[i + 1] - levels[i],
                        facecolor=colors[i],
                        edgecolor="black",
                        linewidth=0.5,
                    )
                )

        for c, p in zip(colors, patches):
            plt.setp(p, "facecolor", c)
            plt.setp(p, "edgecolor", None)

        return patches

    histogram = histogram_legend

    def legend(self, *args, **kwargs):
        if self.legend_type is None:
            return

        try:
            method = getattr(self, self.legend_type)
        except AttributeError:
            raise AttributeError(f"invalid legend type '{self.legend_type}'")

        return method(*args, **kwargs)


class Contour(Style):
    def __init__(
        self,
        *args,
        colors=None,
        line_colors=None,
        labels=False,
        label_kwargs=None,
        interpolate=True,
        **kwargs,
    ):
        super().__init__(*args, colors=colors, **kwargs)
        self._line_colors = line_colors
        self.labels = labels
        self._label_kwargs = label_kwargs or dict()
        self._interpolate = interpolate
        self.kwargs["linewidths"] = kwargs.get("linewidths", 0.5)

    def get_levels(self, layer):
        return layer.mappable.levels

    def plot(self, *args, **kwargs):
        if self._colors is not None:
            if self._interpolate:
                return self.contourf(*args, **kwargs)
            else:
                return self.pcolormesh(*args, **kwargs)
        else:
            return self.contour(*args, **kwargs)

    def to_kwargs(self, data):
        levels = self.levels(data)

        colors = self._colors
        if colors is None:
            colors = self._line_colors or schema.cmap
        colors = expand_colors(colors, levels)

        cmap = LinearSegmentedColormap.from_list(name="", colors=colors, N=len(colors))

        norm = None
        if self.normalize:
            norm = BoundaryNorm(levels, cmap.N)

        return {
            **{"cmap": cmap, "norm": norm, "levels": levels},
            **self.kwargs,
        }

    def to_contour_kwargs(self, data):
        kwargs = super().to_contour_kwargs(data)

        if self._colors and self._line_colors:
            colors = expand_colors(self._line_colors, kwargs["levels"])
            kwargs["cmap"] = LinearSegmentedColormap.from_list(
                name="", colors=colors, N=len(colors)
            )

        return kwargs

    def contourf(self, ax, x, y, values, *args, **kwargs):
        mappable = super().contourf(ax, x, y, values, *args, **kwargs)
        if self._line_colors is not None:
            self.contour(ax, x, y, values, *args, **kwargs)
        return mappable

    def contour(self, *args, **kwargs):
        mappable = super().contour(*args, **kwargs)

        if self.labels:
            self.contour_labels(mappable, **self._label_kwargs)

        return mappable

    def contour_labels(
        self,
        mappable,
        label_fontsize=7,
        label_colors=None,
        label_frequency=1,
        label_background=None,
        label_fmt=None,
    ):
        clabels = mappable.axes.clabel(
            mappable,
            mappable.levels[0::label_frequency],
            inline=True,
            fontsize=label_fontsize,
            colors=label_colors,
            fmt=label_fmt,
            inline_spacing=2,
        )
        if label_background is not None:
            for label in clabels:
                label.set_backgroundcolor(label_background)

        return clabels


class Continuous(Contour):
    def __init__(self, *args, gradients=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.gradients = gradients

    def to_kwargs(self, data):
        levels = self.levels(data)

        colors = self._colors
        if colors is None:
            colors = self._line_colors or schema.cmap
        colors = expand_colors(colors, levels)

        normalised = (levels - np.min(levels)) / (np.max(levels) - np.min(levels))
        color_bins = list(zip(normalised, colors))
        cmap = LinearSegmentedColormap.from_list(name="", colors=color_bins, N=255)

        gradients = self.gradients or [int(255 / len(levels))] * (len(levels) - 1)
        if not isinstance(gradients, (list, tuple)):
            gradients = [gradients] * (len(levels) - 1)

        extrapolated_levels = []
        for i in range(len(levels) - 1):
            bins = list(np.linspace(levels[i], levels[i + 1], gradients[i]))
            extrapolated_levels += bins[(1 if i != 0 else 0) :]
        levels = extrapolated_levels

        norm = None
        if self.normalize:
            norm = BoundaryNorm(levels, cmap.N)

        cmap.set_bad("#D9D9D9", 1.0)

        return {
            **{"cmap": cmap, "norm": norm, "levels": levels},
            **self.kwargs,
        }

    def colorbar(self, *args, ticks=None, **kwargs):
        return super().colorbar(*args, ticks=ticks, **kwargs)


class Hatched(Contour):
    def __init__(self, *args, hatches=".", background_colors=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.hatches = hatches
        self._foreground_colors = self._colors
        self._colors = background_colors or [(0, 0, 0, 0)]

    def contourf(self, *args, **kwargs):
        mappable = super().contourf(*args, hatches=self.hatches, **kwargs)

        colors = expand_colors(self._foreground_colors, mappable.levels)

        for i, collection in enumerate(mappable.collections):
            collection.set_edgecolor(colors[i])
            collection.set_linewidth(0)

        return mappable

    def colorbar(self, *args, **kwargs):
        colorbar = super().colorbar(*args, **kwargs)

        levels = colorbar.mappable.levels

        colors = expand_colors(self._foreground_colors, levels)
        for i, artist in enumerate(colorbar.solids_patches):
            artist.set_edgecolor(colors[i])

        return colorbar

    def disjoint(self, fig, layer, ax, *args, **kwargs):
        legend = super().disjoint(fig, layer, ax, *args, **kwargs)

        colors = expand_colors(self._foreground_colors, layer.mappable.levels)

        for color, artist in zip(colors, legend.get_patches()):
            artist.set_edgecolor(color)
            artist.set_linewidth(0.0)

        return legend


class Scatter(Style):
    def plot(self, *args, **kwargs):
        return self.scatter(*args, **kwargs)

    def get_levels(self, layer):
        return self.levels([layer.mappable.norm.vmin, layer.mappable.norm.vmax])


DEFAULT_STYLE = Style(colors=schema.cmap)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

PRESET_SHAPES = {
    0: (0, 0),
    1: (1, 1),
    2: (1, 2),
    3: (1, 3),
    4: (1, 4),
    5: (2, 3),
    6: (2, 3),
    7: (2, 4),
    8: (2, 4),
    9: (2, 5),
    10: (2, 5),
    11: (3, 4),
    12: (3, 4),
    13: (3, 5),
    14: (3, 5),
    15: (3, 5),
    16: (3, 6),
    17: (3, 6),
    18: (3, 6),
    19: (4, 5),
    20: (4, 5),
}


def rows_cols(num_subplots, rows=None, cols=None, max_cols=8):

    if rows is None and cols is None:
        if num_subplots in PRESET_SHAPES:
            rows, cols = PRESET_SHAPES[num_subplots]
        else:
            cols = max_cols
            rows = num_subplots // max_cols + 1
    elif rows is not None and cols is None:
        if rows == 1:
            cols = num_subplots
        else:
            cols = num_subplots / rows
            if not cols.is_integer():
                cols = int(cols) + 1
            cols = int(cols)
    elif rows is None and cols is not None:
        if cols == 1:
            rows = num_subplots
        else:
            rows = num_subplots / cols
            if not rows.is_integer():
                rows = int(rows) + 1
            rows = int(rows)
    else:
        if rows * cols < num_subplots:
            raise ValueError(
                f"{num_subplots} subplots is too many for a figure with {rows} "
                f"rows and {cols} columns; this figure can contain a maximum "
                f"of {rows*cols} subplots"
            )

    return rows, cols



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

import cartopy.crs as ccrs

GLOBE_AREA = 64800  # 360*180


class CRS:
    def to_ccrs(self, domain):
        return getattr(ccrs, self.cartopy_crs())(**self.get_kwargs(domain))

    def get_kwargs(self, domain):
        attributes = inspect.getmembers(
            self.__class__, lambda attr: not (inspect.isroutine(attr))
        )
        attributes = [
            attr
            for attr in attributes
            if not (attr[0].startswith("__") and attr[0].endswith("__"))
        ]
        return {attr[0].lower(): getattr(domain, attr[1]) for attr in attributes}

    def cartopy_crs(self):
        return self.__class__.__name__


class PlateCarree(CRS):
    CENTRAL_LONGITUDE = "central_lon"


class LambertAzimuthalEqualArea(CRS):
    CENTRAL_LATITUDE = "central_lat"
    CENTRAL_LONGITUDE = "central_lon"


class TransverseMercator(CRS):
    CENTRAL_LATITUDE = "central_lat"
    CENTRAL_LONGITUDE = "central_lon"


class AlbersEqualArea(CRS):
    CENTRAL_LATITUDE = "central_lat"
    CENTRAL_LONGITUDE = "central_lon"
    STANDARD_PARALLELS = "standard_parallels"


class NorthPolarStereo(CRS):
    CENTRAL_LONGITUDE = "central_lon"


class SouthPolarStereo(CRS):
    CENTRAL_LONGITUDE = "central_lon"


class Extents:

    CRS = PlateCarree

    def __init__(self, extents, small_threshold=0.2, large_threshold=0.6):
        self.extents = extents
        self.small_threshold = small_threshold * GLOBE_AREA
        self.large_threshold = large_threshold * GLOBE_AREA

    @property
    def area(self):
        if self.max_lat == self.min_lat:
            multiplicator = 2 * (90 - abs(self.max_lat))
        else:
            multiplicator = self.max_lat - self.min_lat
        return (self.max_lon - self.min_lon) * multiplicator

    @property
    def ratio(self):
        if self.min_lat == self.max_lat:
            denominator = 2 * (90 - abs(self.max_lat))
        else:
            denominator = self.max_lat - self.min_lat
        return (self.max_lon - self.min_lon) / denominator

    @property
    def standard_parallels(self):
        offset = (self.max_lat - self.min_lat) * (4 / 25)
        return (self.min_lat + offset, self.max_lat - offset)

    @property
    def min_lon(self):
        return self.extents[0]

    @property
    def max_lon(self):
        return self.extents[1]

    @property
    def min_lat(self):
        return self.extents[2]

    @property
    def max_lat(self):
        return self.extents[3]

    @property
    def central_lat(self):
        return self.max_lat - (self.max_lat - self.min_lat) / 2

    @property
    def central_lon(self):
        return self.max_lon - (self.max_lon - self.min_lon) / 2

    def is_landscape(self):
        """Landscape domains are at leadt 20% wider than they are tall."""
        return self.ratio > 1.2

    def is_portrait(self):
        """Portrait domains are at least 20% taller than they are wide."""
        return self.ratio < 0.8

    def is_square(self):
        """Square domains have a width and height within 20% of each other."""
        return not self.is_landscape() and not self.is_portrait()

    def is_global(self):
        """Global domains cover > 60% of the globe."""
        return self.area > self.large_threshold

    def is_large(self):
        """Large domains cover < 60% of the globe but > 20%."""
        return not self.is_global() and self.area > self.small_threshold

    def is_small(self):
        """Small domains cover < 20% of the globe."""
        return not self.is_global() and not self.is_large()

    def is_polar(self):
        """Polar domains are centered around a pole."""
        return abs(self.central_lat) > 75

    def is_equatorial(self):
        """Equatorial domains are centered around the equator."""
        return abs(self.central_lat) < 25

    def mutate(self):
        return self

    def get_crs(self):
        return self.CRS().to_ccrs(self)


class Global(Extents):
    """A domain which covers >60% of the globe."""

    CRS = PlateCarree

    def mutate(self):
        if not self.is_global():
            return Equatorial(self.extents)
        return self


class Equatorial(Extents):
    """A domain with a central longitude < ±25 degrees, covering < 20% of the globe."""

    CRS = PlateCarree

    def mutate(self):
        if not self.is_equatorial():
            if self.is_polar():
                return NorthPolar(self.extents)
            else:
                return Square(self.extents)
        elif self.is_large():
            return LargeEqatorial(self.extents)
        return self


class NorthPolar(Extents):
    """A domain with a central longitude > 75 degrees."""

    CRS = NorthPolarStereo

    def mutate(self):
        if self.central_lat < 0:
            return SouthPolar(self.extents)
        return self


class SouthPolar(Extents):
    """A domain with a central longitude < -75 degrees."""

    CRS = SouthPolarStereo


class LargeEqatorial(Extents):
    """A domain with a central longitude < ±25 degrees, covering > 20% of the globe."""

    CRS = TransverseMercator


class Square(Extents):
    """A non-equitorial domain with an aspect ratio > 0.8 and < 1.2."""

    CRS = AlbersEqualArea

    def mutate(self):
        if self.is_landscape():
            return Landscape(self.extents)
        elif self.is_portrait():
            return Portrait(self.extents)
        return self


class Landscape(Extents):
    """A non-equitorial domain with an aspect ratio > 1.2."""

    CRS = AlbersEqualArea


class Portrait(Extents):
    """A non-equitorial domain with an aspect ratio < 0.8."""

    CRS = TransverseMercator


def crs_from_bounds(bounds):
    """
    Get an 'optimal' CRS to use for the given latitude and longitude extents.

    The method for choosing a CRS is as follows:
        - If the area of the map is greater than 60% of the globe, use a global
          equirectangular CRS.
        - If the central latitude is within ±25 degrees, use an equirectangular
          CRS.
        - If the central latitude is greater (less) than +(-)75 degrees, use a
          North (South) Polar Stereo CRS.
        - If the central latitude falls between ±25 and ±75 degrees and the
          aspect ratio is > 0.8, use an Albers Equal Area CRS.
        - If the central latitude falls between ±25 and ±75 degrees and the
          aspect ratio is < 0.8, use a Transverse Mercator CRS.

    This method is adapted from the method used by https://projectionwizard.org,
    which is discussed in the following article:
    Šavrič, B., Jenny, B. and Jenny, H. (2016). Projection Wizard – An online
    map projection selection tool. The Cartographic Journal, 53–2, p. 177–185.
    Doi: 10.1080/00087041.2015.1131938.

    Parameters
    ----------
    extents : list
        The latitude and longitude extents of the bounding box, given as
        `[min_longitude, max_longitude, min_latitude, max_latitude]`.

    Returns
    -------
    cartopy.crs.CRS
    """
    domain = Global(bounds)
    while True:
        new_domain = domain.mutate()
        if new_domain == domain:
            break
        else:
            domain = new_domain
    return domain.get_crs()



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from earthkit.maps import styles

GEOPOTENTIAL_IN_M2_PER_S2 = styles.Contour(
    line_colors="red",
    linewidths=[0.5, 0.5, 0.5, 1],
    labels=True,
    level_step=800,
    units="m2 s-2",
    legend_type=None,
)


GEOPOTENTIAL_HEIGHT_IN_M = styles.Contour(
    line_colors="red",
    linewidths=[0.5, 0.5, 0.5, 1],
    labels=True,
    level_step=50,
    units_orverride="m",
    conversion=lambda x: x / 9.80665,
    legend_type=None,
)


GEOPOTENTIAL_HEIGHT_IN_DAM = styles.Contour(
    line_colors="red",
    linewidths=[0.5, 0.5, 0.5, 1],
    labels=True,
    level_step=5,
    units_override="dam",
    conversion=lambda x: (x / 9.80665) * 0.1,
    legend_type=None,
)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from earthkit.data.utils.bbox import BoundingBox
from earthkit.data.utils.projections import Projection

from earthkit.maps.domains import bounds, crs, optimal
from earthkit.maps.domains.domain import Domain

__all__ = [
    "bounds",
    "crs",
    "optimal",
    "Domain",
]


def parse(domain, crs):
    from earthkit.maps.schemas import schema

    if isinstance(domain, BoundingBox):
        domain = [domain.west, domain.east, domain.south, domain.north]

    if isinstance(crs, Projection):
        crs = crs.to_cartopy_crs()

    if isinstance(domain, str):
        domain = Domain.from_string(domain, crs)
    elif domain is not None or crs is not None:
        domain = Domain(domain, crs)
    else:
        domain = Domain(domain, schema.reference_crs)
    return domain



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from functools import partial

import cartopy
import yaml

from earthkit.maps import definitions

READERS = {
    ".yaml": partial(yaml.load, Loader=yaml.SafeLoader),
}


class DataNotFoundError(FileNotFoundError):
    pass


class AmbiguousDataError(Exception):
    pass


def load(source, data_type=None):
    """
    Load an earthkit.maps auxilliary data file.

    Parameters
    ----------
    source : str
        The name of a file (with or without extension) found within magpye/data.
    data_type : str
        If applicable, the name of the subdirectory withing magpye/data in which
        the auxilliary file will be found.
    """
    path = definitions.DATA_DIR
    if data_type is not None:
        path /= data_type

    matches = list(path.glob(f"{source}*"))
    if not matches:
        raise DataNotFoundError(f"could not find data named '{source}'")
    elif len(matches) > 1:
        raise AmbiguousDataError(
            f"multiple data sources named '{source}'; file extension required"
        )

    path = matches[0]
    reader = READERS.get(path.suffix)
    if reader is None:
        raise KeyError(f"no reader for file with extension {path.suffix}")

    with open(path, "r") as f:
        return reader(f)


def remote_shp(namespace, name, url):
    data_dir = os.path.join(cartopy.config["data_dir"], "shapefiles", namespace)

    file_path = os.path.join(data_dir, f"{name}.shp")

    if not os.path.exists(file_path):
        os.makedirs(data_dir, exist_ok=True)

    return file_path



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from earthkit.data import remote_shp

RESOLUTIONS = {
    "low": "60m",
    "medium": "20m",
    "high": "10m",
    "very high": "03m",
}


EUROSTAT_URL = (
    "https://ec.europa.eu/eurostat/cache/GISCO/distribution/v2/nuts/download/"
    "ref-nuts-{year}-{resolution}.shp.zip"
)


def NUTS(self, level=0, resolution="medium", year=2021):
    """Get NUTS regions."""
    resolution = RESOLUTIONS[resolution]
    url = EUROSTAT_URL.format(year=year, resolution=resolution)
    name = f"nuts-level{level}-{year}-{resolution}"
    path = remote_shp("nuts", name, url)
    return path



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import glob
import importlib

import yaml

from earthkit.maps import definitions, styles
from earthkit.maps.layers.metadata import compare_units


def suggest_style(data, units=None):
    for fname in glob.glob(str(definitions.DATA_DIR / "identities" / "*")):
        with open(fname, "r") as f:
            config = yaml.load(f, Loader=yaml.SafeLoader)
        for criteria in config["criteria"]:
            for key, value in criteria.items():
                if data.metadata(key, default=None) != value:
                    break
            else:
                break
        else:
            continue
        break
    else:
        return styles.DEFAULT_STYLE

    if units is None:
        style = config["styles"]["default"]
    else:
        for unit, style in config["styles"].get("units", {}).items():
            if compare_units(unit, units):
                break
        else:
            raise ValueError(f"{config['id']} has no styles with units {units}")

    module, style = style.split(".")

    module = importlib.import_module(f"earthkit.maps.styles.{module}")

    return getattr(module, style)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections


def recursive_dict_update(original_dict, update_dict):
    """
    Recursively update a dictionary with keys and values from another dictionary.

    Parameters
    ----------
    original_dict : dict
        The original dictionary to be updated (in place).
    update_dict : dict
        The dictionary containing keys to be updated in the original dictionary.
    """
    for k, v in update_dict.items():
        if isinstance(v, collections.abc.Mapping):
            original_dict[k] = recursive_dict_update(original_dict.get(k, {}), v)
        else:
            original_dict[k] = v
    return original_dict



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import yaml

from earthkit.maps import definitions
from earthkit.maps.domains import crs
from earthkit.maps.utils import recursive_dict_update

_DEFAULT_SCHEMA = "default"


class SchemaNotFoundError(FileNotFoundError):
    pass


class _set:
    def __init__(self, schema, **kwargs):
        self.schema = schema

        keys = [key for key in kwargs if key in self.schema]

        self.old_kwargs = {key: self.schema.get(key) for key in keys}
        self.old_kwargs = {
            **self.old_kwargs,
            **{
                key: dict(value)
                for key, value in self.old_kwargs.items()
                if isinstance(value, Schema)
            },
        }

        self.new_kwargs = [key for key in kwargs if key not in self.schema]
        self.schema._update(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.schema._update(**self.old_kwargs)
        for key in self.new_kwargs:
            self.schema.pop(key, None)


class Schema(dict):
    """Class for containing and maintaining global style settings."""

    parsers = {
        "reference_crs": crs.parse,
    }

    def __init__(self, **kwargs):
        self._update(**kwargs)

    def __getattr__(self, key):
        if key in self:
            return self[key]
        raise AttributeError(key)

    def __setattr__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, Schema):
            value = Schema(**value)
        try:
            self[key] = value
        except KeyError:
            raise AttributeError(key)

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"

    def _update(self, **kwargs):
        for key, value in kwargs.items():
            if key in self.parsers:
                value = self.parsers[key](value)
            setattr(self, key, value)

    def apply(self, *keys):
        def decorator(function):
            def wrapper(*args, **kwargs):
                return function(*args, **self._update_kwargs(kwargs, keys))

            return wrapper

        return decorator

    def _update_kwargs(self, kwargs, keys):
        schema_kwargs = self._to_dict()
        if keys:
            schema_kwargs = {key: schema_kwargs[key] for key in keys}
        return recursive_dict_update(schema_kwargs, kwargs)

    def _to_dict(self):
        d = dict()
        for key in self:
            value = getattr(self, key)
            if isinstance(value, type(self)):
                value = value._to_dict()
            d[key] = value
        return d

    def set(self, **kwargs):
        """
        Set the value of a schema key.

        Parameters
        ----------
        **kwargs
            The schema keys and values to set.

        Example
        -------
        >>> schema.set(font="verdana")
        >>> with schema.set(font="comic sans"):
        ...     print(schema.font)
        ...
        comic sans
        >>> print(schema.font)
        verdana
        """
        return _set(self, **kwargs)

    def get(self, key):
        """
        Get the value of a schema key.

        Parameters
        ----------
        key : str
            The name of the schema key to get.

        Example
        -------
        >>> schema.set(font="verdana")
        >>> schema.get("font")
        'verdana'
        """
        return getattr(self, key)

    def use(self, name):
        """
        Use a named schema.

        Parameters
        ----------
        name : str
            The name of the schema to use.

        Example
        -------
        >>> schema.use("default")
        """
        file_name = definitions.SCHEMA_DIR / f"{name}.yaml"
        if not os.path.exists(file_name):
            raise SchemaNotFoundError(f"no schema '{name}' found")
        with open(file_name, "r") as f:
            kwargs = yaml.load(f, Loader=yaml.SafeLoader)
        self._reset(**kwargs)

    def _reset(self, **kwargs):
        self.__init__(**kwargs)


schema = Schema()
schema.use(_DEFAULT_SCHEMA)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cf_units import Unit
from cf_units.tex import tex

PREFERRED_FORMATS = {
    "celsius": "°C",
    "fahrenheit": "°F",
}


def preferred_format(units):
    for name, symbol in PREFERRED_FORMATS.items():
        if compare_units(units, name):
            break
    else:
        symbol = Unit(units).symbol

    try:
        latex = f"${tex(symbol)}$"
    except (SyntaxError, ValueError):
        try:
            latex = f"${tex(symbol.replace('.', ' '))}$"
        except (SyntaxError, ValueError):
            latex = symbol

    return latex


def compare_units(unit_1, unit_2):
    return Unit(unit_1) == Unit(unit_2)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import cv2
import numpy as np

from earthkit.maps import data, domains

DOMAIN_LOOKUP = data.load("domains")


NO_TRANSFORM_FIRST = [
    ccrs.Stereographic,
    ccrs.NearsidePerspective,
]


def force_minus_180_to_180(x):
    return (x + 180) % 360 - 180


def roll_from_0_360_to_minus_180_180(x):
    return np.argwhere(x[0] >= 180)[0][0]


def roll_from_minus_180_180_to_0_360(x):
    return np.argwhere(x[0] >= 0)[0][0]


def force_0_to_360(x):
    return x % 360


class Domain:
    @classmethod
    def from_string(cls, string, crs=None):
        from . import natural_earth

        domain_name = lookup_name(string)

        if crs is not None:
            crs = domains.crs.parse(crs)

        if domain_name is not None and domain_name in DOMAIN_LOOKUP["domains"]:
            domain_config = DOMAIN_LOOKUP["domains"][domain_name]
            if isinstance(domain_config, list):
                bounds = domain_config
                domain_crs = None
            else:
                bounds = domain_config.get("bounds")
                domain_crs = domain_config.get("crs")
            if crs is not None:
                bounds = domains.bounds.from_bbox(
                    bounds, crs, domains.crs.parse(domain_crs)
                )
            else:
                crs = domains.crs.parse(
                    domain_crs or domains.optimal.crs_from_bounds(bounds)
                )
                if domain_crs is None:
                    crs = domains.optimal.crs_from_bounds(bounds)
                    bounds = domains.bounds.from_bbox(bounds, crs)

        else:
            domain_name = domain_name or string
            source = natural_earth.NaturalEarthDomain(domain_name, crs)
            domain_name = source.domain_name
            bounds = source.bounds
            crs = source.crs
        return cls(bounds, crs, domain_name)

    @classmethod
    def from_data(cls, data):
        crs = data.projection().to_cartopy_crs()
        return cls(crs=crs)

    def __init__(self, bounds=None, crs=None, domain_name=None):
        if crs is None:
            if bounds is None:
                self.crs = crs
                self.bounds = bounds
            else:
                self.crs = domains.optimal.crs_from_bounds(bounds)
                self.bounds = domains.bounds.from_bbox(bounds, self.crs)
        else:
            self.crs = domains.crs.parse(crs)
            self.bounds = bounds

        self.domain_name = domain_name

    def __repr__(self):
        return self.title

    @property
    def _can_transform_first(self):
        return not any(isinstance(self.crs, crs) for crs in NO_TRANSFORM_FIRST)

    @property
    def title(self):
        if self.domain_name in DOMAIN_LOOKUP["the_countries"]:
            return f"the {self.domain_name}"
        elif self.domain_name is None:
            if self.bounds is None:
                string = "custom domain"
            else:
                ordinal_values = []
                for lon in self.latlon_bounds[:2]:
                    direction = ""
                    if lon != 0:
                        direction = "°W" if lon > 0 else "°E"
                    ordinal_values.append(f"{round(abs(lon), 2)}{direction}")
                for lat in self.latlon_bounds[2:]:
                    direction = ""
                    if lat != 0:
                        direction = "°N" if lat > 0 else "°S"
                    ordinal_values.append(f"{round(abs(lat), 2)}{direction}")
                string = ", ".join(ordinal_values)
            return string
        return self.domain_name

    @property
    def latlon_bounds(self):
        return domains.bounds.from_bbox(self.bounds, ccrs.PlateCarree(), self.crs)

    def bbox(self, field):
        from earthkit.maps.schemas import schema

        source_crs = field.projection().to_cartopy_crs()

        points = field.to_points(flatten=False)
        values = field.to_numpy(flatten=False)

        if self.bounds and schema.extract_domain:

            try:
                crs_bounds = domains.bounds.from_bbox(self.bounds, source_crs, self.crs)
            except NotImplementedError:
                pass
            else:
                roll_by = None
                if crs_bounds[0] < 0 and crs_bounds[1] > 0:
                    if crs_bounds[0] < points["x"].min():
                        roll_by = roll_from_0_360_to_minus_180_180(points["x"])
                        points["x"] = force_minus_180_to_180(points["x"])
                        for i in range(2):
                            crs_bounds[i] = force_minus_180_to_180(crs_bounds[i])
                elif crs_bounds[0] < 180 and crs_bounds[1] > 180:
                    if crs_bounds[1] > points["x"].max():
                        roll_by = roll_from_minus_180_180_to_0_360(points["x"])
                        points["x"] = force_0_to_360(points["x"])
                        for i in range(2):
                            crs_bounds[i] = force_0_to_360(crs_bounds[i])
                if roll_by is not None:
                    points["x"] = np.roll(points["x"], roll_by, axis=1)
                    points["y"] = np.roll(points["y"], roll_by, axis=1)
                    values = np.roll(values, roll_by, axis=1)

                bbox = np.where(
                    (points["x"] >= crs_bounds[0])
                    & (points["x"] <= crs_bounds[1])
                    & (points["y"] >= crs_bounds[2])
                    & (points["y"] <= crs_bounds[3]),
                    True,
                    False,
                )

                kernel = np.ones((8, 8), dtype="uint8")
                bbox = cv2.dilate(bbox.astype("uint8"), kernel).astype(bool)

                shape = bbox[np.ix_(np.any(bbox, axis=1), np.any(bbox, axis=0))].shape

                points["x"] = points["x"][bbox].reshape(shape)
                points["y"] = points["y"][bbox].reshape(shape)
                values = values[bbox].reshape(shape)

        return values, points


def lookup_name(domain_name):
    """Format a domain name."""
    # normalise the input string and the lookup key
    domain_name = domain_name.lower().replace("_", " ")
    name_mapping = {k.lower(): k for k in DOMAIN_LOOKUP["domains"]}

    if domain_name not in name_mapping:
        for name, alt_names in DOMAIN_LOOKUP["alternate_names"].items():
            if domain_name in [alt_name.lower() for alt_name in alt_names]:
                domain_name = name
                break
        else:
            domain_name = None
    domain_name = name_mapping.get(domain_name, domain_name)

    return domain_name



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt
import numpy as np

# from earthkit.maps.metadata import labels

DEFAULT_COLORBAR_TITLE = "{variable_name} ({units})"


def colorbar(
    layer,
    title=DEFAULT_COLORBAR_TITLE,
    orientation=None,
    location=None,
    format="g",
    **kwargs,
):
    if orientation and location:
        raise ValueError(
            "'orientation' and 'location' are mutually exclusive arguments"
        )
    elif orientation is not None:
        location = "right" if orientation == "vertical" else "bottom"
    elif location is None:
        # Default to a colorbar on the right
        location = "right"
    orientation = "vertical" if location in ("left", "right") else "horizontal"

    auto = False
    if not any(key in kwargs for key in ("cax", "x", "y")):
        kwargs["cax"] = layer._chart.fig.add_axes([0, 0, 0, 0])
        auto = True

    if format is not None:
        kwargs["format"] = lambda x, _: f"{x:{format}}"

    try:
        levels = layer.layer.norm.boundaries
    except AttributeError:
        levels = layer.layer.levels
    if "ticks" not in kwargs and len(np.unique(np.ediff1d(levels))) != 1:
        kwargs["ticks"] = levels

    cbar = plt.colorbar(layer.layer, orientation=orientation, **kwargs)

    # if title:
    #     rotation = {"right": 270, "left": 90, "top": 0, "bottom": 0}[location]
    #     labelpad = {"right": 20, "left": -60, "top": -60, "bottom": 0}[location]
    #     cbar.set_label(
    #         labels.title_from_layer(layer, title),
    #         rotation=rotation,
    #         labelpad=labelpad,
    #     )

    cbar.auto = auto  # indicate whether the colorbar location is automatic

    cbar.location = location

    return cbar



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from string import Formatter

import numpy as np
from cf_units import Unit
from cf_units.tex import tex


class BaseFormatter(Formatter):

    SUBPLOT_ATTRIBUTES = {
        "domain": "domain",
        "crs": "crs_name",
    }

    STYLE_ATTRIBUTES = {
        "units": "units",
    }

    def convert_field(self, value, conversion):
        if conversion == "u":
            return str(value).upper()
        elif conversion == "l":
            return str(value).lower()
        return super().convert_field(value, conversion)

    def format_keys(self, format_string, kwargs):
        keys = (i[1] for i in self.parse(format_string) if i[1] is not None)
        for key in keys:
            kwargs[key] = self.format_key(key)
        return kwargs

    def format_key(self, key):
        return key

    def format(self, format_string, /, *args, **kwargs):
        kwargs = self.format_keys(format_string, kwargs)
        return super().format(format_string, *args, **kwargs)


DEFAULT_FORECAST_TITLE = (
    "{variable_name}\n"
    "Base time: {base_time:%H:%M} on {base_time:%Y-%m-%d}   "
    "Valid time: {valid_time:%H:%M} on {valid_time:%Y-%m-%d} (T+{lead_time})"
)
DEFAULT_ANALYSIS_TITLE = (
    "{variable_name} at {valid_time:%H:%M} on {valid_time:%Y-%m-%d}"
)

TIME_KEYS = ["base_time", "valid_time", "lead_time", "time"]

MAGIC_KEYS = {
    "variable_name": {
        "preference": ["long_name", "standard_name", "name", "short_name"],
    }
}


PRETTY_UNITS = {
    "celsius": "°C",
    "fahrenheit": "°F",
}


def default_label(data):
    if data.metadata("type") == "an":
        format_string = DEFAULT_ANALYSIS_TITLE
    else:
        format_string = DEFAULT_FORECAST_TITLE
    return format_string


def compare_units(unit_1, unit_2):
    from cf_units import Unit

    return Unit(unit_1) == Unit(unit_2)


def format_units(units):
    for name, formatted_units in PRETTY_UNITS.items():
        try:
            if compare_units(units, name):
                units = formatted_units
                break
        except ValueError:
            continue
    else:
        try:
            units = str(Unit(units))
        except ValueError:
            pass
    return f"${tex(units)}$"


def get_metadata(layer, attr):

    if attr in TIME_KEYS:
        handler = TimeHandler(layer.data.datetime())
        label = getattr(handler, attr)[0]

    else:
        candidates = [attr]
        if attr in MAGIC_KEYS:
            candidates = MAGIC_KEYS[attr]["preference"] + candidates

        for item in candidates:
            label = layer.data.metadata(item, default=None)
            if label is not None:
                break
        else:
            raise KeyError(f"No key {attr} found in metadata")

    return label


class TimeHandler:
    def __init__(self, times):
        if not isinstance(times, (list, tuple)):
            times = [times]
        for i, time in enumerate(times):
            if not isinstance(time, dict):
                times[i] = {"time": time}
        self.times = times

    def _extract_time(method):
        def wrapper(self):
            attr = method.__name__
            times = [self._named_time(time, attr) for time in self.times]
            _, indices = np.unique(times, return_index=True)
            return [times[i] for i in sorted(indices)]

        return property(wrapper)

    @staticmethod
    def _named_time(time, attr):
        return time.get(attr, time.get("time"))

    @property
    def time(self):
        return self.valid_time

    @_extract_time
    def base_time(self):
        pass

    @_extract_time
    def valid_time(self):
        pass

    @property
    def lead_time(self):
        if len(self.base_time) == 1 and len(self.valid_time) > 1:
            times = itertools.product(self.base_time, self.valid_time)
        elif len(self.base_time) == len(self.valid_time):
            times = zip(self.base_time, self.valid_time)
        else:
            times = [
                (
                    self._named_time(time, "base_time"),
                    self._named_time(time, "valid_time"),
                )
                for time in self.times
            ]
        return [int((vtime - btime).total_seconds() / 3600) for btime, vtime in times]


def list_to_human(iterable, conjunction="and", oxford_comma=False):
    """Convert a list to a human-readable string."""
    list_of_strs = [str(item) for item in iterable]

    if len(list_of_strs) > 2:
        list_of_strs = [", ".join(list_of_strs[:-1]), list_of_strs[-1]]
        if oxford_comma:
            list_of_strs[0] += ","

    return f" {conjunction} ".join(list_of_strs)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from earthkit.maps import styles

TEMPERATURE_AT_2M_IN_CELSIUS = styles.Contour(
    colors="turbo",
    extend="both",
    levels=range(-40, 41, 2),
    units="celsius",
)


TEMPERATURE_AT_2M_IN_KELVIN = styles.Contour(
    colors="turbo",
    extend="both",
    levels=range(230, 311, 2),
    units="K",
)


TEMPERATURE_AT_2M_IN_FAHRENHEIT = styles.Contour(
    colors="turbo",
    extend="both",
    levels=range(-40, 101, 5),
    units="fahrenheit",
)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from earthkit.maps import Superplot

DEFAULT_BLUEPRINT = {
    "coastlines": {},
    "borders": {},
    "gridlines": {},
    "legend": {},
    "subplot_titles": {},
}


def quickplot(*data, blueprint=DEFAULT_BLUEPRINT, units=None, **kwargs):
    chart = Superplot(**kwargs)

    if units is not None:
        if not isinstance(units, (list, tuple)):
            units = [units]
        if len(units) != len(data):
            raise ValueError(
                f"units must be speficied for each input data; got {len(data)} "
                f"data fields but {len(units)} units"
            )
    else:
        units = [None] * len(data)

    for item, unit in zip(data, units):
        chart.plot(item, units=unit)

    for method, method_kwargs in blueprint.items():
        getattr(chart, method)(**method_kwargs)

    chart.show()



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    # NOTE: the `version.py` file must not be present in the git repository
    #   as it is generated by setuptools at install time
    from .version import __version__
except ImportError:  # pragma: no cover
    # Local copy or not installed with setuptools
    __version__ = "999"

import glob
import os

from matplotlib import font_manager, rcParams

from earthkit.maps import styles
from earthkit.maps.definitions import FONTS_DIR
from earthkit.maps.layers.superplots import Superplot
from earthkit.maps.quickplot import quickplot
from earthkit.maps.schemas import schema

__all__ = [
    "quickplot",
    "styles",
    "Superplot",
    "schema",
]


def register_fonts():
    fontpaths = glob.glob(os.path.join(FONTS_DIR, "*"))
    for fontpath in fontpaths:
        font_files = glob.glob(os.path.join(fontpath, "*.ttf"))
        for font_file in font_files:
            font_manager.fontManager.addfont(font_file)


register_fonts()
rcParams["font.family"] = schema.font
rcParams["axes.linewidth"] = 0.5
rcParams["axes.edgecolor"] = "#555"



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import earthkit.data
import matplotlib.pyplot as plt

from earthkit.maps import domains, layouts
from earthkit.maps.layers import metadata
from earthkit.maps.layers.layers import MultiplotLayer
from earthkit.maps.layers.subplots import Subplot, SubplotFormatter
from earthkit.maps.schemas import schema


class SuperplotFormatter(metadata.BaseFormatter):
    def __init__(self, subplots, unique=True):
        self.subplots = subplots
        self.unique = unique
        self._layer_index = None

    def convert_field(self, value, conversion):
        f = super().convert_field
        if isinstance(value, list):
            return [f(v, conversion) for v in value]
        else:
            return f(value, conversion)

    def format_key(self, key):
        values = [
            SubplotFormatter(subplot).format_key(key) for subplot in self.subplots
        ]
        values = [item for sublist in values for item in sublist]
        return values

    def format_field(self, value, format_spec):
        f = super().format_field
        if isinstance(value, list):
            values = [f(v, format_spec) for v in value]
            if self._layer_index is not None:
                value = values[self._layer_index]
                self._layer_index = None
            else:
                if self.unique:
                    values = list(dict.fromkeys(values))
                value = metadata.list_to_human(values)
        return value


class Superplot:

    MAX_COLS = 8

    @classmethod
    def from_gridspec(cls, gridspec, *args, **kwargs):
        obj = cls(*args, **kwargs)
        obj._gridspec = gridspec
        return obj

    def __init__(self, domain=None, domain_crs=None, crs=None, rows=None, cols=None):

        if domain_crs is not None:
            domain_crs = domains.crs.parse(domain_crs)
            domain = domains.bounds.from_bbox(domain, crs, domain_crs)

        self._domain = domain
        self._crs = crs

        self.domain = domains.parse(domain, crs)

        self._fig = None
        self._gridspec = None

        self._rows = rows
        self._cols = cols

        self.subplots = []
        self._subplots_generator = None

        self._queue = []

    def __len__(self):
        return len(self.subplots)

    def __getitem__(self, i):
        while i + 1 > len(self.subplots):
            self.add_subplot()
        return self.subplots[i]

    @property
    def fig(self):
        if self._fig is None:
            self._fig = plt.figure(
                figsize=schema.figsize, constrained_layout=True, dpi=schema.dpi
            )
        return self._fig

    @property
    def gridspec(self):
        if self._gridspec is None:
            self._gridspec = self.fig.add_gridspec(*self.shape)
        return self._gridspec

    @property
    def rows(self):
        if self._rows is None:
            if len(self) == 0:
                raise ValueError("cannot get rows from empty figure")
            self._rows = layouts.rows_cols(
                len(self), cols=self._cols, max_cols=self.MAX_COLS
            )[0]
        return self._rows

    @property
    def cols(self):
        if self._cols is None:
            if len(self) == 0:
                raise ValueError("cannot get cols from empty figure")
            self._cols = layouts.rows_cols(
                len(self), rows=self._rows, max_cols=self.MAX_COLS
            )[1]
        return self._cols

    @property
    def shape(self):
        return self.rows, self.cols

    @property
    def subplots_generator(self):
        if self._subplots_generator is None:
            self._subplots_generator = (
                (row, col)
                for row, col in itertools.product(
                    range(self.gridspec.nrows),
                    range(self.gridspec.ncols),
                )
            )
        return self._subplots_generator

    def add_subplot(self, *args, data=None, domain=None, crs=None, **kwargs):
        row, col = next(self.subplots_generator)

        if domain is None and crs is None:
            domain = self._domain
            crs = self._crs

        if not args:
            args = (self.gridspec[row, col],)

        if data is not None:
            subplot = Subplot.from_data(
                self, data, *args, domain=domain, crs=crs, **kwargs
            )
        else:
            subplot = Subplot(self, *args, domain=domain, crs=crs, **kwargs)

        self.subplots.append(subplot)
        return subplot

    @property
    def distinct_legend_layers(self):
        """Group layers by style."""
        subplot_layers = [subplot.distinct_legend_layers for subplot in self.subplots]
        subplot_layers = [item for sublist in subplot_layers for item in sublist]

        groups = []
        for layer in subplot_layers:
            for i in range(len(groups)):
                if groups[i][0].style == layer.style:
                    groups[i].append(layer)
                    break
            else:
                groups.append([layer])

        groups = [MultiplotLayer(layers) for layers in list(groups)]

        return groups

    def defer(method):
        """Defer a method's execution until this superplot has subplots."""

        def wrapper(self, *args, **kwargs):
            if not self.subplots:
                self._queue.append((method, args, kwargs))
            else:
                return method(self, *args, **kwargs)

        return wrapper

    def expand_rows_cols(method):
        def wrapper(self, data, *args, **kwargs):
            if not hasattr(data, "__len__") or not isinstance(
                data, earthkit.data.core.Base
            ):
                data = [data]

            if not self.subplots:
                num_subplots = len(data)
                self._rows, self._cols = layouts.rows_cols(
                    num_subplots,
                    rows=self._rows,
                    cols=self._cols,
                    max_cols=self.MAX_COLS,
                )

            layers = []
            for i, field in enumerate(data):
                if i + 1 > len(self.subplots):
                    subplot = self.add_subplot(data=field)
                else:
                    subplot = self.subplots[i]
                layer = getattr(subplot, method.__name__)(field, *args, **kwargs)
                layers.append(layer)

            return MultiplotLayer(layers)

        return wrapper

    @expand_rows_cols
    def plot(self, *args, **kwargs):
        pass

    @expand_rows_cols
    def contourf(self, *args, **kwargs):
        pass

    shaded_contour = contourf

    @expand_rows_cols
    def contour(self, *args, **kwargs):
        pass

    @expand_rows_cols
    def pcolormesh(self, *args, **kwargs):
        pass

    @expand_rows_cols
    def scatter(self, *args, **kwargs):
        pass

    @defer
    def coastlines(self, *args, **kwargs):
        return [subplot.coastlines(*args, **kwargs) for subplot in self.subplots]

    @defer
    def borders(self, *args, **kwargs):
        return [subplot.borders(*args, **kwargs) for subplot in self.subplots]

    @defer
    def land(self, *args, **kwargs):
        return [subplot.land(*args, **kwargs) for subplot in self.subplots]

    @defer
    def ocean(self, *args, **kwargs):
        return [subplot.ocean(*args, **kwargs) for subplot in self.subplots]

    @defer
    def stock_img(self, *args, **kwargs):
        return [subplot.stock_img(*args, **kwargs) for subplot in self.subplots]

    @defer
    def gridlines(self, *args, **kwargs):
        return [subplot.gridlines(*args, **kwargs) for subplot in self.subplots]

    def format_string(self, string, unique=True, grouped=True):
        if not grouped:
            results = [
                subplot.format_string(string, unique) for subplot in self.subplots
            ]
            result = metadata.list_to_human(results)
        else:
            result = SuperplotFormatter(self.subplots, unique=unique).format(string)
        return result

    def _release_queue(self):
        while len(self._queue):
            method, args, kwargs = self._queue.pop(0)
            method(self, *args, **kwargs)

    @schema.legend.apply()
    def legend(self, *args, fontsize=None, location=None, **kwargs):
        legends = []
        for i, layer in enumerate(self.distinct_legend_layers):
            if isinstance(location, (list, tuple)):
                loc = location[i]
            else:
                loc = location
            if layer.style is not None:
                legend = layer.style.legend(
                    self.fig, layer, *args, ax=layer.axes, location=loc, **kwargs
                )
            legends.append(legend)
        if len(legends) == 1:
            legends = legends[0]
        return legends

    @property
    def _default_title_template(self):
        return self.subplots[0]._default_title_template

    @schema.title.apply()
    def title(self, label=None, unique=True, grouped=True, wrap=True, **kwargs):
        if label is None:
            label = self._default_title_template
        label = self.format_string(label, unique, grouped)
        return self.fig.suptitle(label, wrap=wrap, **kwargs)

    def subplot_titles(self, *args, **kwargs):
        return [subplot.title(*args, **kwargs) for subplot in self.subplots]

    def show(self, *args, **kwargs):
        """Display the chart."""
        self._release_queue()
        plt.show(*args, **kwargs)

    def save(self, *args, **kwargs):
        """Save the chart."""
        self._release_queue()
        plt.savefig(*args, **kwargs)
